'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var BEGIN = 'BEGIN';
var COMMIT = 'COMMIT';
var REVERT = 'REVERT';
var INITIAL_OPTIMIST = {};

module.exports = optimist;
module.exports.BEGIN = BEGIN;
module.exports.COMMIT = COMMIT;
module.exports.REVERT = REVERT;
function optimist(fn) {
  return function (state, action) {
    var _ref = state || {};

    var _ref$optimist = _ref.optimist;
    var optimist = _ref$optimist === undefined ? INITIAL_OPTIMIST : _ref$optimist;

    var oldState = _objectWithoutProperties(_ref, ['optimist']);

    var oldOptimist = optimist;
    if (!state) oldState = undefined;
    if (action.optimist && (action.optimist.type === COMMIT || action.optimist.type === REVERT)) {
      var _optimist = optimist;
      var transaction = _optimist[action.optimist.id];

      var transactions = _objectWithoutProperties(_optimist, [action.optimist.id]);

      if (!transaction) {
        console.error('Cannot ' + action.optimist.type + ' transaction with id "' + action.optimist.id + '" because it does not exist');
      }
      optimist = transactions;
      if (transaction && action.optimist.type === REVERT) {
        (function () {
          var state = transaction.state;
          var actions = transaction.actions;

          actions.forEach(function (action) {
            state = fn(state, action);
          });
          oldState = state;
        })();
      }
    }
    if (Object.keys(optimist).length) {
      (function () {
        var newOptimist = {};
        Object.keys(optimist).forEach(function (key) {
          newOptimist[key] = { state: optimist[key].state, actions: optimist[key].actions.concat([action]) };
        });
        optimist = newOptimist;
      })();
    }
    if (action.optimist && action.optimist.type === BEGIN) {
      if (action.optimist.id in optimist) {
        console.error('Implicitly committing transaction with id "' + action.optimist.id + '" because it already exists, and you are starting' + ' it again.');
      }
      if (!state) {
        console.error('You should never begin an optimistic transaction before initializing your store.' + ' You would have nothing to revert to!');
      }
      optimist = _extends({}, optimist, _defineProperty({}, action.optimist.id, { state: oldState, actions: [] }));
    }
    var newState = fn(oldState, action);
    if (!newState || typeof newState !== 'object' || Array.isArray(newState)) {
      throw new TypeError('Error while handling "' + action.type + '": Optimist requires that state is always a plain object.');
    }
    if (oldOptimist !== optimist || !equal(newState, oldState)) return _extends({ optimist: optimist }, newState);else return state;
  };
}

function equal(newState, oldState) {
  if (newState === oldState) return true;
  if (!(newState && oldState)) return false;
  for (var key in newState) {
    if (newState[key] !== oldState[key]) {
      return false;
    }
  }
  for (var key in oldState) {
    if (newState[key] !== oldState[key]) {
      return false;
    }
  }
  return true;
}